from datasets import load_dataset
import pandas as pd
import pytest


@pytest.fixture(scope="session")
def data():
    dataset = "mteb/mtop_domain"
    dset = load_dataset(dataset)
    return pd.DataFrame(dset["test"])[["text", "label", "label_text"]]



import pandas as pd
import string
import random
from typing import Tuple, List, Union
from pandas._typing import RandomState


class _DataProcessor:
    """Encapsulates data preparation steps pre-curation
    and processing results post-curation."""

    def __init__(self, random_state: RandomState):
        self.random_state = random_state

        self.row_id_col = None

    def _generate_random_suffix(self, length: int = 3) -> str:
        """
        Generates and returns a unique string with letters and/or digits.

        Args:
            length (int): Length of the string (default is 3)

        Returns:
            str: Unique short string
        """
        # Define the characters to use in the string
        characters = string.ascii_letters + string.digits

        # Generate a random string of the specified length
        random_string = "".join(random.choices(characters, k=length))

        return random_string

    def _create_col_name(
        self, df_columns: List[Union[str, int]], base_col_name: str
    ) -> str:
        """Returns a valid non conflicting column name based on
        the input `base_col_name`

        Args:
            df_columns (List[Union[str, int]]): List of column names in the dataframe object
            base_col_name (str): base string to which suffix will be added
        Returns:
            str: new column name
        """
        new_col_name = base_col_name
        while new_col_name in df_columns:
            new_col_name += f"_{self._generate_random_suffix()}"

        return new_col_name

    def _add_row_id(self, df: pd.DataFrame) -> Tuple[pd.DataFrame, str]:
        """Adds a new column representing row numbers in input data

        Args:
            pd.DataFrame: Input data
        Returns:
            pd.DataFrame: Data with integer labels
            str: Name of the column added
        """
        # Generate a unique column name
        row_id_col = self._create_col_name(df.columns, base_col_name="row_id")

        # Add row numbers as a new column
        df[row_id_col] = range(len(df))

        # Save name of the new column for downstream usage
        self.row_id_col = row_id_col

        return df, row_id_col

    def _shuffle_data(self, df: pd.DataFrame) -> pd.DataFrame:
        """Shuffles the data using pandas sampling

        Args:
            pd.DataFrame: Input data

            Returns:
                pd.DataFrame: Shuffled data
        """
        return df.sample(frac=1.0, random_state=self.random_state).reset_index(
            drop=True
        )

    def _convert_labels_to_int(
        self, df: pd.DataFrame, y_col_name: str
    ) -> Tuple[pd.DataFrame, str]:
        """Map string labels to integers and add them as a new column in the data
        for downstream processing

        Args:
            df (pd.DataFrame): Data with labels
        Returns:
            pd.DataFrame: Data with integer labels
            str: newly added column name for integer labels
        """

        if not df[y_col_name].dtype == "object":
            return df, y_col_name

        # Generate a unique column name
        y_col_name_int = self._create_col_name(df.columns, base_col_name="label_int")

        # Create a mapping dictionary from string values to integers
        unique_labels = df[y_col_name].unique()
        label_mapping = {text: i for i, text in enumerate(unique_labels)}

        # Map string values to integers and create a new column
        df[y_col_name_int] = df[y_col_name].map(label_mapping)

        return df, y_col_name_int

    def _preprocess(
        self, df: pd.DataFrame, y_col_name: str
    ) -> Tuple[pd.DataFrame, str, str]:
        """
        Adds the following columns to the DataFrame -
        1) 'label_int' with integer values for all labels
        2) row numbers if random_state is not None

        Args:
            pd.DataFrame: Data to be processed
            y_col_name: label column in the data
        Returns:
            pd.DataFrame: Data with added columns
            str: Name of the newly added column
        """

        df, y_col_name_int = self._convert_labels_to_int(df, y_col_name)

        df, row_id_col = self._add_row_id(df)
        df = self._shuffle_data(df)

        return df, row_id_col, y_col_name_int

    def _postprocess(self, df: pd.DataFrame, display_cols: List[str]) -> pd.DataFrame:
        """Removes redundant columns and returns the data
        with selected columns to display

        Args:
            df (pd.DataFrame): Data to be processed
            display_cols (List[str]): Non redundant columns in data

        Returns:
            pd.DataFrame: Data after post-processing data
        """

        row_id_col = self.row_id_col

        if not row_id_col:
            return df[display_cols]

        return df.sort_values(by=[row_id_col]).reset_index(drop=True)[display_cols]



from sklearn.pipeline import Pipeline
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import StratifiedKFold
from typing import Generator, Tuple
import pandas as pd


def _supported_sklearn_artifacts(**options) -> Tuple[dict, dict]:
    """Returns collection of supported artifacts for curation representation and model

    Returns:
        Tuple[dict, dict]: two dictionaries - one for curation representation and one for curation model
    """
    representation_dict = {
        "tfidf": TfidfVectorizer(
            analyzer=options.get("analyzer", "word"),
            ngram_range=options.get("ngram_range", (1, 1)),
            max_features=min(
                options["num_samples"] // 10, options.get("max_features", 1000)
            ),
        )
    }

    model_dict = {"logistic_regression": LogisticRegression()}

    return representation_dict, model_dict


def _get_pipeline(curate_representation: str, curate_model: str, **options) -> Pipeline:
    """Returns the pipeline for a given curation representation and model choice

    Args:
        curate_representation (str): Feature representation method to be used during curation.
        curate_model (str): Machine learning model that is trained with
                     `representation` features during curation.

    Raises:
        ValueError: If a given `curate_representation` or `curate_model` is not supported

    Returns:
        Pipeline: the constructed sklearn pipeline
    """

    representation_dict, model_dict = _supported_sklearn_artifacts(**options)

    representation_artifact = representation_dict.get(curate_representation)
    model_artifact = model_dict.get(curate_model)

    if not representation_artifact:
        raise ValueError(
            f"{curate_representation} is not supported. Please select `representation='tfidf'` instead "
        )

    if not model_artifact:
        raise ValueError(
            f"{model_artifact} is not supported. Please select `model='logistic_regression'` instead "
        )

    return Pipeline(
        [
            (curate_representation, representation_artifact),
            (curate_model, model_artifact),
        ]
    )


def _data_splitter(
    df: pd.DataFrame,
    X_col_name: str,
    y_col_name_int: str,
    strategy: str = "stratified_kfold",
    n_splits: int = 5,
) -> Generator:
    """Splits the data into train and validation to be consumed in CrossValCurate

    Args:
        df (pd.DataFrame): Input data
        X_col_name (str): Column to be used to extract input features
        y_col_name_int (str): Label column in data
        strategy (str, optional): Strategy to use while splitting the data. Defaults to 'stratified_kfold'.
        n_splits (int, optional): The number of splits. Defaults to 5.

    Returns:
        StratifiedKFold: Currently, a StratifiedKFold object.
    """

    if strategy == "stratified_kfold":
        cv = StratifiedKFold(n_splits=n_splits, shuffle=False, random_state=None)
        return cv.split(df[X_col_name], df[y_col_name_int])

    raise ValueError(
        f"Data splitting strategy {strategy} not supported. Please use `strategy='stratified_kfold'`"
    )



from ._generic import (
    Logger,
    _is_valid,
    _check_supported,
    add_asymmetric_noise,
    _exception_handler,
)
from ._sklearn_artifacts import _get_pipeline, _data_splitter
from ._dataprocessing import _DataProcessor

__all__ = ["Logger", "add_asymmetric_noise"]



import pytest
from dqc import CrossValCurate
from dqc.utils import add_asymmetric_noise
import pandas as pd


@pytest.mark.parametrize("curate_representation", ["tfidf", "TfIdf"])
@pytest.mark.parametrize(
    "curate_model",
    ["logistic_regression", "Logistic_Regression"],
)
@pytest.mark.parametrize("calibration_method", [None, "calibrate_using_baseline"])
@pytest.mark.parametrize("random_state", [None, 1])
def test_crossvalcurate_success(
    data, curate_representation, curate_model, calibration_method, random_state
):
    cvc = CrossValCurate(
        calibration_method=calibration_method, random_state=random_state
    )
    data_curated = cvc.fit_transform(data)

    # Check if dataframe is return
    assert type(data_curated) == pd.DataFrame

    # Check column presence
    assert all(
        col in data_curated.columns
        for col in [
            "prediction_probability",
            "predicted_label",
            "label_correctness_score",
            "is_label_correct",
        ]
    )

    # Check data types
    assert all(
        data_curated[col].dtype == dtype
        for col, dtype in [
            ("prediction_probability", float),
            ("label_correctness_score", float),
            ("is_label_correct", bool),
        ]
    )

    # Check non-null values
    assert all(
        data_curated[col].notnull().all()
        for col in [
            "prediction_probability",
            "predicted_label",
            "label_correctness_score",
            "is_label_correct",
        ]
    )


@pytest.mark.parametrize("curate_representation", ["tf_idf", 1, ""])
@pytest.mark.parametrize(
    "curate_model",
    ["logistic regression", 2, ""],
)
@pytest.mark.parametrize("calibration_method", ["random", 3, ""])
@pytest.mark.parametrize("random_state", ["randomstr", ""])
def test_crossvalcurate_failure(
    data, curate_representation, curate_model, calibration_method, random_state
):
    with pytest.raises(ValueError):
        cvc = CrossValCurate(
            calibration_method=calibration_method, random_state=random_state
        )
        data = cvc.fit_transform(data)


def test_crossvalcurate_datafailure(data):
    with pytest.raises(ValueError):
        cvc = CrossValCurate()
        data = data.groupby("label", group_keys=False).sample(
            frac=0.005, random_state=43
        )
        data = cvc.fit_transform(data)


@pytest.mark.parametrize("calibration_method", [None, "calibrate_using_baseline"])
@pytest.mark.parametrize("verbose", [True, False])
def test_verbosity(data, calibration_method, verbose):
    cvc = CrossValCurate(calibration_method=calibration_method, verbose=verbose)
    data = cvc.fit_transform(data)


@pytest.mark.parametrize("noise_prob", [0.1])
@pytest.mark.parametrize("random_state", [None, 1])
def test_add_asymmetric_noise(data, noise_prob, random_state):
    noisy_labels = add_asymmetric_noise(
        data["label"], noise_prob=noise_prob, random_state=random_state
    )

    assert isinstance(noisy_labels, pd.Series)
    assert len(noisy_labels) == len(data["label"])



from dqc.base import BaseCurate
from dqc.utils import (
    Logger,
    _get_pipeline,
    _is_valid,
    _DataProcessor,
    _data_splitter,
    _exception_handler,
)

from sklearn.preprocessing import MinMaxScaler
from tqdm import tqdm
import numpy as np
import pandas as pd
from typing import Union, List, Tuple

logger = Logger("C2")

tqdm.pandas()


class CrossValCurate(BaseCurate):
    """

    Args:
        n_splits (int, optional): Number of splits to use when running cross-validation based curation.
        verbose (bool, optional): Sets the verbosity level during execution. `True` indicates logging level INFO and `False` indicates logging level 'ERROR'.

    """

    def __init__(self, n_splits: int = 5, verbose: bool = False, **options):
        super().__init__(**options)
        self.n_splits = n_splits
        self.verbose = verbose
        self._set_verbosity(verbose)

        self.curate_pipeline = None
        self.scaler = None
        self.y_col_name_int = None
        self.result_col_list = [
            "label_correctness_score",
            "is_label_correct",
            "predicted_label",
            "prediction_probability",
        ]

    def __str__(self):
        display_dict = self.__dict__.copy()
        for key in list(display_dict.keys()):
            if key in [
                "curate_pipeline",
                "scaler",
                "y_col_name_int",
                "result_col_list",
            ]:
                ## Don't need to display these attributes
                del display_dict[key]

        return str(display_dict)

    __repr__ = __str__

    def _is_confident(self, row: pd.Series) -> bool:
        """Return a boolean variable indicating whether we are confident
           about the correctness of the label assigned to a given data sample

        Args:
            row (pd.Series): data sample whose label correctness need to be evaluated

        Returns:
            bool: `True` if we are confident that the label assigned is correct
                else `False`
        """
        threshold = self.correctness_threshold
        if (row["predicted_label"] == row[self.y_col_name_int]) and (
            row["label_correctness_score"] > threshold
        ):
            return True
        return False

    def _set_verbosity(self, verbose: bool):
        """Set logger level based on user input for parameter `verbose`

        Args:
            verbose (bool): Indicator for verbosity
        """
        if verbose:
            logger.set_level("INFO")
        else:
            logger.set_level("WARNING")

    def _no_calibration(
        self,
        input_labels: List[Union[int, str]],
        pred_prob_matrix: np.ndarray,
        label_list: List[Union[int, str]],
    ) -> Tuple[List[Union[int, str]], List[float]]:
        """Returns predictions, corresponding probabilities and label correctness
        scores without any calibration

        Args:
            input_labels (List[Union[int, str]]): _description_
            pred_prob_matrix (np.ndarray): _description_
            label_list (List[Union[int, str]]): _description_

        Returns:
            Tuple[List[Union[int, str]], List[float]]: _description_
        """
        pred_probs = np.max(pred_prob_matrix, axis=1).tolist()
        preds = [label_list[index] for index in np.argmax(pred_prob_matrix, axis=1)]
        label_correctness_scores = [
            pred_prob_matrix[row_index, label_list.index(label)]
            for row_index, label in enumerate(input_labels)
        ]

        return preds, pred_probs, label_correctness_scores

    def _get_baselines(self, data_with_noisy_labels: pd.DataFrame) -> pd.DataFrame:
        """Computes the baseline prediction probabilities using
          input label distribution

        Args:
            data_with_noisy_labels (pd.DataFrame): Input data with
                                            corresponding noisy labels

        Returns:
            pd.DataFrame: Labels and corresponding probabilities
        """
        thresholds_df = (
            data_with_noisy_labels[self.y_col_name_int].value_counts(1)
        ).reset_index()
        thresholds_df.columns = [self.y_col_name_int, "probability"]
        return thresholds_df

    def _calibrate_using_baseline(
        self,
        input_labels: List[Union[int, str]],
        pred_prob_matrix: np.ndarray,
        label_list: List[Union[int, str]],
        baseline_probs: pd.DataFrame,
    ) -> Tuple[List[Union[int, str]], List[float], List[float]]:
        """Calibrate the predicted probabilities using baseline probabilities and
        returns the modified predictions, corresponding probabilities and label
        correctness scores

        Args:
            input_labels (List[Union[int, str]]): Noisy labels from data
            pred_prob_matrix (np.ndarray): Predicted probabilities per label
                                for each sample in the data
            label_list (List[Union[int, str]]): Ordered list of labels where
                                ordering matches `pred_prob_matrix`
            baseline_probs (pd.DataFrame): Prediction probabilities computed
                                using input label distribution

        Returns:
            Tuple[List[Union[int, str]], List[float], List[float]]:
            Returns the following
            'calibrated_prediction' : Label predictions post calibration
            'calibrated_probabilities' : Normalized scores corresponding
                                    to 'calibrated_prediction'
            'label_correctness_score' : Normalized scores corresponding
                                     to 'label' (provided as input to
                                    `self.fit_transform`)
        """

        label_list_df = pd.DataFrame({"label_list": label_list})
        baseline_probs = pd.merge(
            label_list_df,
            baseline_probs,
            left_on="label_list",
            right_on=self.y_col_name_int,
            how="inner",
        ).reset_index(drop=True)

        baseline_probs.drop("label_list", axis=1, inplace=True)

        prob_array = baseline_probs["probability"].values
        pred_prob_matrix = (pred_prob_matrix - prob_array) / prob_array

        if not self.scaler:
            self.scaler = MinMaxScaler()

        pred_prob_matrix = self.scaler.fit_transform(pred_prob_matrix)

        calibrated_predictions = [
            label_list[index] for index in np.argmax(pred_prob_matrix, axis=1)
        ]
        calibrated_probabilities = np.max(pred_prob_matrix, axis=1).tolist()
        label_correctness_scores = [
            pred_prob_matrix[row_index, label_list.index(label)]
            for row_index, label in enumerate(input_labels)
        ]

        return (
            calibrated_predictions,
            calibrated_probabilities,
            label_correctness_scores,
        )

    @_exception_handler
    def fit_transform(
        self,
        data_with_noisy_labels: pd.DataFrame,
        X_col_name: str = "text",
        y_col_name: str = "label",
        **options,
    ) -> pd.DataFrame:
        """Fit CrossValCurate on the given input data

        Args:
            data_with_noisy_labels (pd.DataFrame): Input data with corresponding noisy labels
            X_col_name (str): Column to be used to extract input features.
            y_col_name (str): Label column in the data.
        Returns:

            pd.DataFrame: Input Data samples with \n
                1) Corresponding predicted_labels using CrossValCurate \n
                2) Prediction probabilities \n
                3) An indicator variable `is_label_correct` capturing label correctness
        """
        _is_valid(data_with_noisy_labels, X_col_name, y_col_name)
        n_splits = self.n_splits
        options["num_samples"] = len(data_with_noisy_labels)

        logger.info("Pre-processing the data..")
        dp = _DataProcessor(
            random_state=self.random_state,
        )

        data_with_noisy_labels, row_id_col, y_col_name_int = dp._preprocess(
            data_with_noisy_labels, y_col_name=y_col_name
        )

        # y_col_name_int needs to be accessed downstream
        self.y_col_name_int = y_col_name_int

        data_columns = [X_col_name, y_col_name_int]
        logger.info(
            f"Building the curation pipeline with {n_splits}-fold cross validation.."
        )
        self.curate_pipeline = _get_pipeline(
            self.curate_representation, self.curate_model, **options
        )

        split_indices = _data_splitter(
            data_with_noisy_labels,
            X_col_name=X_col_name,
            y_col_name_int=y_col_name_int,
            n_splits=self.n_splits,
        )

        # Lists to store predictions
        predictions = []
        prediction_probabilities = []
        label_correctness_scores = []

        # Keep track of shuffled data
        row_ids = []

        if self.calibration_method == "calibrate_using_baseline":
            logger.info("Computing baseline predictions for each label..")
            baseline_probs = self._get_baselines(data_with_noisy_labels[data_columns])

        # Iterate through kfold splits
        for train_index, val_index in tqdm(split_indices):
            # Split the data
            X_train, X_val = (
                data_with_noisy_labels.loc[train_index, X_col_name].values,
                data_with_noisy_labels.loc[val_index, X_col_name].values,
            )
            y_train, y_val = (
                data_with_noisy_labels.loc[train_index, y_col_name_int].values,
                data_with_noisy_labels.loc[val_index, y_col_name_int].values,
            )

            # Train the model
            self.curate_pipeline.fit(X_train, y_train)
            classes_ = self.curate_pipeline.classes_.tolist()

            # Make predictions on the validation set
            y_pred_probs = self.curate_pipeline.predict_proba(X_val)

            if self.calibration_method == "calibrate_using_baseline":
                y_preds, y_pred_probs, label_cscores = self._calibrate_using_baseline(
                    y_val,
                    y_pred_probs,
                    label_list=classes_,
                    baseline_probs=baseline_probs,
                )
            else:
                y_preds, y_pred_probs, label_cscores = self._no_calibration(
                    y_val, y_pred_probs, label_list=classes_
                )

            predictions.extend(y_preds)
            prediction_probabilities.extend(y_pred_probs)
            label_correctness_scores.extend(label_cscores)
            row_ids.extend(data_with_noisy_labels.loc[val_index, row_id_col].values)

        # Order dataframe according to `rowids``

        row_id_df = pd.DataFrame()
        row_id_df[row_id_col] = pd.Series(row_ids)
        data_with_noisy_labels = pd.merge(
            row_id_df, data_with_noisy_labels, how="left", on=row_id_col
        )

        # Add results as columns
        data_with_noisy_labels["label_correctness_score"] = pd.Series(
            label_correctness_scores
        )
        data_with_noisy_labels["predicted_label"] = pd.Series(predictions)
        data_with_noisy_labels["prediction_probability"] = pd.Series(
            prediction_probabilities
        )

        logger.info("Identifying the correctly labelled samples..")
        data_with_noisy_labels["is_label_correct"] = (
            data_with_noisy_labels.progress_apply(self._is_confident, axis=1)
        )

        return dp._postprocess(
            data_with_noisy_labels, display_cols=data_columns + self.result_col_list
        )



import logging
import pandas as pd
import numpy as np
import string
import random

from typing import Tuple, List, Union
from pandas._typing import RandomState
from functools import wraps


class Logger:
    """Logging utility"""

    levels = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]

    def __init__(self, name, verbosity_level="error"):
        self.logger = logging.getLogger(name)
        self.set_level(verbosity_level)

        self._setup_stream_handler()
        self.logger.propagate = False

    def _setup_stream_handler(self):
        console_handler = logging.StreamHandler()
        formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
        console_handler.setFormatter(formatter)
        self.logger.addHandler(console_handler)

    def set_level(self, verbosity_level):
        if verbosity_level in self.levels:
            self.logger.setLevel(verbosity_level)

    def get_level(self):
        return self.logger.getEffectiveLevel()

    def info(self, message):
        self.logger.info(f"{message}")

    def warning(self, message):
        self.logger.warning(f"{message}")


def _check_num_unique_labels(data: pd.DataFrame, y_col_name: str):
    """Check if there are atleast two labels in the input data

    Args:
        data (pd.DataFrame): Input data with noisy labels
        y_col_name (str): Label column in data

    Raises:
        ValueError: if number of labels <= 1
    """
    count = len(pd.unique(data[y_col_name]))
    if count <= 1:
        raise ValueError(f"Number of distinct labels should be > 1 " f"Found {count}")


def _check_columns(data: pd.DataFrame, X_col_name: str, y_col_name: str):
    """Sanity checks for column names in input data

    Args:
        data (pd.DataFrame): Input data with noisy labels
        X_col_name (str): Column to be used to extract input features
        y_col_name (str): Label column in data

    Raises:
        ValueError: If any of the expected columns are not present in data
    """
    actual_columns = data.columns.tolist()

    if not X_col_name:
        raise ValueError("'X_col_name' cannot be None. Please pass a valid column name")

    if not y_col_name:
        raise ValueError("'y_col_name' cannot be None ")

    expected_columns = [X_col_name, y_col_name]

    if set(expected_columns) > set(actual_columns):
        raise ValueError(
            f"Data does not contain the expected columns. "
            f"Expected(X_col_name, y_col_name): {expected_columns}, "
            f"Actual: {actual_columns}"
        )


def _check_null_values(data: pd.DataFrame, X_col_name: str, y_col_name: str):
    """Sanity checks to detect null values in data

    Args:
        data (pd.DataFrame): Input data with noisy labels
        X_col_name (str): Column to be used to extract input features
        y_col_name (str): Label column in data

    Raises:
        ValueError: If null values are found in the data
    """
    if any(data[col].isnull().any() for col in [X_col_name, y_col_name]):
        raise ValueError(
            "Null values found in the data. \
                    Automatically imputing missing values is not supported yet."
        )


def _is_valid(data: pd.DataFrame, X_col_name: str, y_col_name: str):
    """Run collection of sanity checks on input data

    Args:
        data (pd.DataFrame): Input data with noisy labels
        X_col_name (str): Column to be used to extract input features
        y_col_name (str): Label column in data
    """
    _check_columns(data, X_col_name, y_col_name)
    _check_num_unique_labels(data, y_col_name)
    _check_null_values(data, X_col_name, y_col_name)


def _fetch_supported_implementations():
    """Returns the list of valid options that can be passed as values
    for parameters in BaseCurate

    Returns:
        dict: Mapping of parameters and their corresponding list of options
    """
    return {
        "curate_representation": ["tfidf"],
        "curate_model": ["logistic_regression"],
        "calibration_method": [None, "calibrate_using_baseline"],
    }


def _check_supported(
    curate_representation: str,
    curate_model: str,
    calibration_method: Union[str, None],
):
    """Checks if any of the input parameter values are invalid

    Args:
        curate_representation (str, optional): Feature representation method
                                    to be used during curation.
        curate_model (str): Machine learning model to be used during curation.
        calibration_method (Union[str, None]): Approach to be used for calibration
                                     of `curate_model` predictions

    Raises:
        ValueError: If any of the input parameter values are invalid
    """
    supported = _fetch_supported_implementations()
    msg = ""
    if curate_representation not in supported["curate_representation"]:
        msg += f"curate_representation '{curate_representation}' is not supported. "
        msg += (
            f"Currently, {', '.join(supported['curate_representation'])} is supported\n"
        )

    if curate_model not in supported["curate_model"]:
        msg += f"curate_model '{curate_model}' is not supported. "
        msg += f"Currently, {', '.join(supported['curate_model'])} is supported\n"

    if calibration_method not in supported["calibration_method"]:
        msg += f"calibration_method '{calibration_method}' is not supported. "
        msg += f"Currently, {', '.join(map(str, supported['calibration_method']))} are supported\n"

    if len(msg) > 0:
        raise ValueError(msg)


def add_asymmetric_noise(
    labels: pd.Series,
    noise_prob: float = 0.3,
    random_state: Union[RandomState, None] = None,
):
    """
    Util function to add asymmetric noise to labels
    for simulation of noisy label scenarios.

    Args:
        labels (pd.Series): Input pandas Series with integer values
                        ranging from 0 to n - 1.
        noise_prob (float): Probability of adding noise to each value.
        random_state (Union[RandomState, None]): Random seed for reproducibility
    Returns:
        pd.Series: Series with asymmetric noise added to it.
    """
    # Set seed
    np.random.seed(random_state)

    # Avoid modifying the original data
    noisy_labels = labels.copy()

    # Assuming labels are in the range 0 to n, get n
    max_value = noisy_labels.max()

    # Determine the number of samples to modify based on the noise ratio
    num_samples = min(len(noisy_labels), int(len(noisy_labels) * noise_prob + 1))

    # Sample random indices from the labels to introduce noise
    target_indices = np.random.choice(len(noisy_labels), num_samples, replace=False)

    for idx in target_indices:
        # Generate random noise for the value in this position
        noise = np.random.choice([-1, 1], size=1, p=[noise_prob, 1 - noise_prob])

        # Add noise to the series values
        noisy_labels[idx] += noise

        # Ensure values remain within the valid range [0, max_value]
        noisy_labels[idx] = noisy_labels[idx] % max_value

    return noisy_labels


def _exception_handler(func):
    def wrapper(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except ValueError as ve:
            if "cannot be greater than the number of members in each class" in str(ve):
                raise ValueError(
                    f"Not enough data samples per label with n_splits={self.n_splits} in CrossValCurate."
                )
            raise

    return wrapper



from abc import ABC, abstractmethod
from dqc.utils import _check_supported
from typing import Union
from pandas._typing import RandomState


class BaseCurate(ABC):
    """Base class for data curation to compute label correctness scores
      and identify reliable labelled samples
    Args:
        curate_representation (str, optional): Feature representation method to be used during curation.
        Defaults to 'tfidf' (`sklearn.feature_extraction.text.TfidfVectorizer`).
        curate_model (str, optional): Machine learning model that is trained with
                     `curate_representation` features during curation.
        Defaults to 'logistic_regression' (sklearn.linear_model.LogisticRegression).
        calibration_method (Union[str, None], optional): Approach to be used for calibration
                    of `curate_model` predictions. Defaults to None.
        correctness_threshold (float, optional): Minimum prediction probability using
          `curate_model` to consider the corresponding sample as 'correctly labelled'.
          Defaults to 0.8.
        random_state (RandomState, optional): Random seed for
                                reproducibility. Defaults to None.
    """

    def __init__(
        self,
        curate_representation: str = "tfidf",
        curate_model: str = "logistic_regression",
        calibration_method: Union[str, None] = "calibrate_using_baseline",
        correctness_threshold: float = 0.5,
        random_state: RandomState = 42,
        **options,
    ):
        self.curate_representation = curate_representation.lower()
        self.curate_model = curate_model.lower()
        self.calibration_method = calibration_method

        _check_supported(curate_representation, curate_model, calibration_method)

        self.correctness_threshold = correctness_threshold
        self.random_state = random_state

    @abstractmethod
    def fit_transform(self): ...
